from graph_tool import Graph
import itertools
import sys
import typing

from kgtk.io.kgtkreader import KgtkReader

def load_graph_from_kgtk(kr: KgtkReader,
                         directed: bool=False,
                         inverted: bool=False,
                         eprop_types: typing.Optional[typing.List[str]]=None,
                         hashed: bool=True,
                         hash_type: str="string", # for future support
                         ecols: typing.Optional[typing.Tuple[int, int]]=None,
                         pcol: typing.Optional[int]=None,
                         pset: typing.Optional[typing.Set[str]]=None,
                         ipset: typing.Optional[typing.Set[str]]=None,
                         upset: typing.Optional[typing.Set[str]]=None,
                         out: typing.TextIO = sys.stderr,
                         verbose: bool = False,
                         ):
    """Load a graph from a `KgtkReader` file containing a list of edges and edge
    properties.  This code is based on load_graph_from_csv(...) in
    `graph-tool/src/graph_tool/__init__.py`, downloaded from git.skewed.de on
    27-Jul-2020.

    Parameters
    ----------
    kr : ``KgtkReader``

    directed : ``bool`` (optional, default: ``False``)
        Whether or not the graph is directed.

    inverted : ``bool`` (optional, default: ``False``)
        Whether or not the links are inverted.

    eprop_types : list of ``str`` (optional, default: ``None``)
        List of edge property types to be read from remaining columns (if this
        is ``None``, all properties will be of type ``string``.

    hashed : ``bool`` (optional, default: ``True``)
        If ``True`` the vertex values in the edge list are not assumed to
        correspond to vertex indices directly. In this case they will be mapped
        to vertex indices according to the order in which they are encountered,
        and a vertex property map with the vertex values is returned.

    hash_type : ``str`` (optional, default: ``string``)
        If ``hashed == True``, this will determined the type of the vertex values.
        It can be any property map value type (see :class:`PropertyMap`).

        Note: As of 29-Jul-2020, this parameter to graph.add_edge_list(...) is
        supported in the git version of graph_tools, but is not mentioned in
        the graph-tools 2.33 documentation.

    ecols : pair of ``int`` (optional, default: ``(0,1)``)
        Line columns used as source and target for the edges.

    pcol : ``int`` (optional, default: ``None``)
        Line column ised as predicate filter with pset.

    pset : set of `str` (optional, default: ``None``)
        When ``pcol`` and ``pset`` are both supplied, the input edges
        will be filtered to inlude only edges (rows) with predicate
        (label) values in ``pset``.

    ipset : set of `str` (optional, default: ``None``)
        When ``pcol``, and ``ipset`` are both supplied, input edges
        (rows) predicate (label) values in ``pset`` will have their
        source and target columns swapped.

    upset : set of `str` (optional, default: ``None``)
        When ``pcol``, and ``ipset`` are both supplied, input edges
        (rows) predicate (label) values in ``pset`` will be duplicated with their
        source and target columns swapped.

    Returns
    -------
    g : :class:`~graph_tool.Graph`
        The loaded graph. It will contain additional columns in the file as
        internal edge property maps. If ``hashed == True``, it will also contain
        an internal vertex property map with the vertex names.

    """
    if ecols is None:
        ecols = (kr.node1_column_idx, kr.node2_column_idx)

    r = kr # R may be wrapped for column reordering and/or non-hashed use.

    if pcol is not None and pset is not None and len(pset) > 0:
        def filter_rows(rows):
            for row in rows:
                if row[pcol] in pset:
                    yield row
        r = filter_rows(r)

    if pcol is not None and ipset is not None and len(ipset) > 0:
        def invert_specific_rows(rows):
            for row in rows:
                if row[pcol] in ipset:
                    row = list(row)
                    row[ecols[0]], row[ecols[1]] = row[ecols[1]], row[ecols[0]]
                    yield row
                else:
                    yield row
        r = invert_specific_rows(r)

    if pcol is not None and upset is not None and len(upset) > 0:
        def duplicate_and_invert_specific_rows(rows):
            for row in rows:
                yield row
                if row[pcol] in upset:
                    row = list(row)
                    row[ecols[0]], row[ecols[1]] = row[ecols[1]], row[ecols[0]]
                    yield row
        r = duplicate_and_invert_specific_rows(r)

    if ecols != (0, 1):
        def reorder(rows):
            for row in rows:
                row = list(row)
                s = row[ecols[0]]
                t = row[ecols[1]]
                del row[min(ecols)]
                del row[max(ecols)-1]
                yield [s, t] + row
        r = reorder(r)

    if inverted:
        def invert(rows):
            for row in rows:
                row = list(row)
                row[0], row[1] = row[1], row[0]
                yield row
        r = invert(r)

    if not hashed:
        def conv(rows):
            for row in rows:
                row = list(row)
                row[0] = int(row[0])
                row[1] = int(row[1])
                yield row
        r = conv(r)

    g = Graph(directed=directed)

    if eprop_types is None:
        if verbose:
            print("eprop_types is None", file=out, flush=True)
        eprops = [g.new_ep("string") for x in kr.column_names[2:]]
    else:
        if verbose:
            print("eprop_types: [%s]" % (", ".join([repr(x) for x in eprop_types])), file=out, flush=True)
        if len(eprop_types) != kr.column_count - 2:
            raise ValueError("There are %d eprop columns and %d eprop types." % (kr.column_count - 2, len(eprop_types)))
        eprops = [g.new_ep(t) for t in eprop_types]

    # 29-Jul-2020: This is supported in the git.skewed.de repository, and
    # presumably will be supported in a future release.  Unfortunately, graph-tool
    # doesn't appear to include a version indicator or API version indicator
    # easily accessible from Python.
    #
    # The graph-tool 2.33 documentation does not include the hash_type parameter.
    #
    # name = g.add_edge_list(itertools.chain([line], r), hashed=hashed,
    #                       hash_type=hash_type, eprops=eprops)
    if verbose:
        print("Adding edges from the input file.", file=out, flush=True)
    name = g.add_edge_list(r, hashed=hashed, eprops=eprops)
    if verbose:
        print("Done adding edges from the input file.", file=out, flush=True)
    g.vp.name = name

    eprop_names: typing.List[str] = list(kr.column_names)
    del eprop_names[min(ecols)]
    del eprop_names[max(ecols)-1]
    if verbose:
        print("eprop_names: [%s]" % (", ".join([repr(x) for x in eprop_names])), file=out, flush=True)

    for i, p in enumerate(eprops):
        ename = eprop_names[i]
        g.ep[ename] = p
        if verbose:
            print("prop %d name=%s" % (i, repr(ename)), file=out, flush=True)

    return g



from enum import Enum

class ValidationAction(Enum):
    """
    What should we when a validation check fails?

    TODO: Why can't this be inside class KgtkReader?
    """
    PASS = 0 # Silently ignore the validation problem and continue to process the line.
    REPORT = 1 # Report the validation problem and continue to process the line.
    EXCLUDE = 2 # Silently exclude the line from further processing.
    COMPLAIN = 3 # Report the validation problem and exclude the line from further processing.
    ERROR = 4 # Report the validation problem and immediately raise a ValueError.
    EXIT = 5 # Report the validation problem and immediately exit





"""Replace symbol names to move relationships form one KG to another..

"""

from argparse import Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Replace item and relationship identifiers in a KGTK file.',
        'description': 'Replace item and relationship values to move a network from one symbol set to another. ' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert replace-nodes --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file(positional=True)
    parser.add_output_file()

    parser.add_input_file(who="A KGTK file with mapping records",
                          dest="mapping_file",
                          options=["--mapping-file"],
                          optional=False)

    parser.add_output_file(who="A KGTK output file that will contain unmodified edges.",
                           dest="unmodified_edges_file",
                           options=["--unmodified-edges-file"],
                           metavar="UNMODIFIED_EDGES_FILE",
                           optional=True)

    parser.add_output_file(who="A KGTK output file that will contain activated mapping edges.",
                           dest="activated_mapping_file",
                           options=["--activated-mapping-edges-file"],
                           metavar="ACTIVATED_MAPPING_EDGES_FILE",
                           optional=True)

    parser.add_output_file(who="A KGTK output file that will contain rejected mapping edges.",
                           dest="rejected_mapping_file",
                           options=["--rejected-mapping-edges-file"],
                           metavar="REJECTED_MAPPING_EDGES_FILE",
                           optional=True)

    parser.add_argument(      "--confidence-column", dest="confidence_column_name",
                              help=h("The name of the confidence column.  (default=%(default)s)"),
                              default="confidence")

    parser.add_argument(      "--require-confidence", dest="require_confidence",
                              help=h("If true, require a confidence column with non-empty values. (default=%(default)s)."),
                              metavar="True/False",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--default-confidence-value", dest="default_confidence_str",
                              help=h("The default confidence value when the confidence column is missing " +
                                     "or a mapping edge does not have a confidence value. (default=None)"))
    
    parser.add_argument(      "--threshold", dest="confidence_threshold",
                              help="The minimum acceptable confidence value. Mapping records with a lower" +
                              " confidence value are excluded. (default=%(default)f)",
                              type=float, default=1.0)
    
    parser.add_argument(      "--same-as-item-label", dest="same_as_item_label",
                              help=h("The name of the mapping property for mapping the node1 and node2 columns.  (default=%(default)s)"),
                              default="same_as_item")

    parser.add_argument(      "--same-as-property-label", dest="same_as_property_label",
                              help=h("The name of the mapping property for mapping the label column.  (default=%(default)s)"),
                              default="same_as_property")

    parser.add_argument(      "--allow-exact-duplicates", dest="allow_exact_duplicates",
                              help=h("When True, allow duplicate mapping entries with the same node2 values.  (default=%(default)s)"),
                              metavar="True/False",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--allow-idempotent-mapping", dest="allow_idempotent_mapping",
                              help=h("When True, allow mapping entries having node1 == node2. Otherwise, filter them out.  (default=%(default)s)"),
                              metavar="True/False",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--split-output-mode", dest="split_output_mode",
                              help="If true, send only modified edges to the output file. (default=%(default)s).",
                              metavar="True/False",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--modified-pattern", dest="modified_pattern",
                              help=h("A pattern that defines a significant edge modifications.  (default=%(default)s)"),
                              default="node1|label|node2")

    parser.add_argument(      "--node1-column", dest="node1_column_name",
                              help=h("The name of the node1 column in the input file.  (default=node1 or its alias)"))

    parser.add_argument(      "--label-column", dest="label_column_name",
                              help=h("The name of the label column in the input file.  (default=label or its alias)"))

    parser.add_argument(      "--node2-column", dest="node2_column_name",
                              help=h("The name of the node2 column in the input file.  (default=node2 or its alias)"))

    parser.add_argument(      "--mapping-rule-mode", dest="mapping_rule_mode",
                              choices=["normal", "same-as-item", "same-as-property"],
                              help=h("Force a mapping rule mode.  (default=%(default)s"),
                              default="normal")

    parser.add_argument(      "--mapping-node1-column", dest="mapping_node1_column_name",
                              help=h("The name of the node1 column in the mapping file.  (default=node1 or its alias)"))

    parser.add_argument(      "--mapping-label-column", dest="mapping_label_column_name",
                              help=h("The name of the label column in the mapping file.  (default=label or its alias)"))

    parser.add_argument(      "--mapping-node2-column", dest="mapping_node2_column_name",
                              help=h("The name of the node2 column in the mapping file.  (default=node2 or its alias)"))

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode[parsed_shared_args._mode],
                                    expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who="input", expert=_expert, defaults=False)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who="mapping", expert=_expert, defaults=False)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        mapping_file: KGTKFiles,
        unmodified_edges_file: KGTKFiles,
        activated_mapping_file: KGTKFiles,
        rejected_mapping_file: KGTKFiles,

        confidence_column_name: str,
        require_confidence: bool,
        default_confidence_str: typing.Optional[str],
        confidence_threshold: float,

        same_as_item_label: str,
        same_as_property_label: str,
        allow_exact_duplicates: bool,
        allow_idempotent_mapping: bool,

        split_output_mode: bool,
        modified_pattern: str,

        node1_column_name: typing.Optional[str],
        label_column_name: typing.Optional[str],
        node2_column_name: typing.Optional[str],
        mapping_rule_mode: str,
        mapping_node1_column_name: typing.Optional[str],
        mapping_label_column_name: typing.Optional[str],
        mapping_node2_column_name: typing.Optional[str],

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from pathlib import Path
    import sys
    
    from kgtk.exceptions import KGTKException
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)
    mapping_kgtk_file: Path = KGTKArgumentParser.get_input_file(mapping_file, who="KGTK mappping file")
    unmodified_edges_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(unmodified_edges_file, who="KGTK unmodified edges output file")
    activated_mapping_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(activated_mapping_file, who="KGTK activated mapping output file")
    rejected_mapping_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(rejected_mapping_file, who="KGTK rejected mapping output file")

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    input_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs, who="input", fallback=True)
    mapping_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs, who="mapping", fallback=True)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % repr(str(input_kgtk_file)), file=error_file, flush=True)
        print("--output-file=%s" % repr(str(output_kgtk_file)), file=error_file, flush=True)
        print("--mapping-file=%s" % repr(str(mapping_kgtk_file)), file=error_file, flush=True)
        if unmodified_edges_kgtk_file is not None:
            print("--unmodified-edges-file=%s" % repr(str(unmodified_edges_kgtk_file)), file=error_file, flush=True)
        if activated_mapping_kgtk_file is not None:
            print("--activated-mapping-edges-file=%s" % repr(str(activated_mapping_kgtk_file)), file=error_file, flush=True)
        if rejected_mapping_kgtk_file is not None:
            print("--rejected-mapping-edges-file=%s" % repr(str(rejected_mapping_kgtk_file)), file=error_file, flush=True)

        print("--confidence-column=%s" % repr(confidence_column_name), file=error_file, flush=True)
        print("--require-confidence=%s" % repr(require_confidence), file=error_file, flush=True)
        if default_confidence_str is not None:
            print("--default-confidence-value=%s" % default_confidence_str, file=error_file, flush=True)
        print("--threshold=%f" % confidence_threshold, file=error_file, flush=True)

        print("--same-as-item-label=%s" % repr(same_as_item_label), file=error_file, flush=True)
        print("--same-as-property-label=%s" % repr(same_as_property_label), file=error_file, flush=True)
        print("--allow-exact-duplicates=%s" % repr(allow_exact_duplicates), file=error_file, flush=True)
        print("--allow-idempotent-actions=%s" % repr(allow_idempotent_mapping), file=error_file, flush=True)

        print("--split-output-mode=%s" % repr(split_output_mode), file=error_file, flush=True)
        print("--modified-pattern=%s" % repr(modified_pattern), file=error_file, flush=True)

        if node1_column_name is not None:
            print("--node1-column-=%s" % repr(node1_column_name), file=error_file, flush=True)
        if label_column_name is not None:
            print("--label-column-=%s" % repr(label_column_name), file=error_file, flush=True)
        if node2_column_name is not None:
            print("--node2-column-=%s" % repr(node2_column_name), file=error_file, flush=True)
        print("--mapping-rule-mode=%s" % repr(mapping_rule_mode), file=error_file, flush=True)
        if mapping_node1_column_name is not None:
            print("--mapping-node1-column-=%s" % repr(mapping_node1_column_name), file=error_file, flush=True)
        if mapping_label_column_name is not None:
            print("--mapping-label-column-=%s" % repr(mapping_label_column_name), file=error_file, flush=True)
        if mapping_node2_column_name is not None:
            print("--mapping-node2-column-=%s" % repr(mapping_node2_column_name), file=error_file, flush=True)

        input_reader_options.show(out=error_file, who="input")
        mapping_reader_options.show(out=error_file, who="mapping")
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    default_confidence_value: typing.Optional[float] = None
    if default_confidence_str is not None:
        try:
            default_confidence_value = float(default_confidence_str)
        except:
            raise KGTKException("--default-confidence-value=%s is invalid" % repr(default_confidence_str))

    try:

        if verbose:
            print("Opening the mapping file %s." % repr(str(mapping_kgtk_file)), file=error_file, flush=True)
        mkr:  KgtkReader = KgtkReader.open(mapping_kgtk_file,
                                           options=mapping_reader_options,
                                           value_options = value_options,
                                           error_file=error_file,
                                           verbose=verbose,
                                           very_verbose=very_verbose,
        )
        trouble = False
        mapping_node1_idx: int = mkr.get_node1_column_index(mapping_node1_column_name)
        mapping_label_idx: int = mkr.get_label_column_index(mapping_label_column_name)
        mapping_node2_idx: int = mkr.get_node2_column_index(mapping_node2_column_name)
        if mapping_node1_idx < 0:
            trouble = True
            print("Error: Cannot find the mapping file node1 column.", file=error_file, flush=True)
        if mapping_label_idx < 0 and mapping_rule_mode == "normal":
            trouble = True
            print("Error: Cannot find the mapping file label column.", file=error_file, flush=True)
        if mapping_node2_idx < 0:
            trouble = True
            print("Error: Cannot find the mapping file node2 column.", file=error_file, flush=True)
        if trouble:
            # Clean up:                                                                                                                                               
            mkr.close()
            raise KGTKException("Missing columns in the mapping file.")
        confidence_column_idx: int = mkr.column_name_map.get(confidence_column_name, -1)
        if require_confidence and confidence_column_idx < 0:
            mkr.close()
            raise KGTKException("The mapping file does not have a confidence column, and confidence is required.")
        
        rmkw: typing.Optional[KgtkWriter] = None
        if rejected_mapping_kgtk_file is not None:
            if verbose:
                print("Opening the rejected mapping edges file %s." % repr(str(rejected_mapping_kgtk_file)), file=error_file, flush=True)
            rmkw = KgtkWriter.open(mkr.column_names,
                                   rejected_mapping_kgtk_file,
                                   mode=KgtkWriter.Mode[mkr.mode.name],
                                   use_mgzip=input_reader_options.use_mgzip, # Hack!
                                   mgzip_threads=input_reader_options.mgzip_threads, # Hack!
                                   error_file=error_file,
                                   verbose=verbose,
                                   very_verbose=very_verbose)

        # Mapping structures:
        item_map: typing.MutableMapping[str, str] = dict()
        item_line_map: typing.MutableMapping[str, int] = dict()
        property_map: typing.MutableMapping[str, str] = dict()
        property_line_map: typing.MutableMapping[str, int] = dict()

        mapping_rows: typing.MutableMapping[int, typing.List[str]] = dict()
        activated_mapping_rows: typing.MutableMapping[int, typing.List[str]] = dict()

        # Read the mapping file.
        if verbose:
            print("Processing the mapping file.", file=error_file, flush=True)
        mapping_confidence_exclusions: int = 0
        mapping_idempotent_exclusions: int = 0
        mapping_errors: int = 0
        mapping_line_number: int = 0
        mrow: typing.List[str]
        for mrow in mkr:
            mapping_line_number += 1
            mapping_node1: str = mrow[mapping_node1_idx]
            mapping_label: str = mrow[mapping_label_idx] if mapping_rule_mode == "normal" else ""
            mapping_node2: str = mrow[mapping_node2_idx]
            mapping_confidence: typing.Optional[float] = default_confidence_value
            if confidence_column_idx >= 0:
                confidence_value_str: str = mrow[confidence_column_idx]
                if len(confidence_value_str) == 0:
                    if require_confidence:
                        print("In line %d of the mapping file: the required confidence value is missing" % (mapping_line_number),
                              file=error_file, flush=True)
                        mapping_errors += 1
                        continue
                else:
                    try:
                        mapping_confidence = float(confidence_value_str)
                    except ValueError:
                        print("In line %d of the mapping file: cannot parse confidence value %s" % (mapping_line_number, repr(mrow[confidence_column_idx])),
                              file=error_file, flush=True)
                        mapping_errors += 1
                        continue
            if mapping_confidence is not None and mapping_confidence < confidence_threshold:
                mapping_confidence_exclusions += 1
                if rmkw is not None:
                    rmkw.write(mrow)
                continue

            if mapping_node1 == mapping_node2 and not allow_idempotent_mapping:
                mapping_idempotent_exclusions += 1
                continue
        
            if mapping_rule_mode == "same-as-item" or mapping_label == same_as_item_label:
                if mapping_node1 in item_map:
                    if mapping_node2 != item_map[mapping_node1] or not allow_exact_duplicates:
                        print("Duplicate %s for %s at mapping file line %d, originally in line %d" % (mapping_label,
                                                                                                      repr(mapping_node1),
                                                                                                      mapping_line_number,
                                                                                                      item_line_map[mapping_node1]),
                              file=error_file, flush=True)
                        mapping_errors += 1
                    continue

                item_map[mapping_node1] = mapping_node2
                item_line_map[mapping_node1] = mapping_line_number
                mapping_rows[mapping_line_number] = mrow.copy()

            elif mapping_rule_mode == "same-as-property" or mapping_label == same_as_property_label:
                if mapping_node1 in property_map:
                    if mapping_node2 != property_map[mapping_node1] or not allow_exact_duplicates:
                        print("Duplicate %s for %s at mapping file line %d, originally in line %d" % (mapping_label,
                                                                                                      repr(mapping_node1),
                                                                                                      mapping_line_number,
                                                                                                      property_line_map[mapping_node1]),
                              file=error_file, flush=True)
                        mapping_errors += 1
                    continue
                property_map[mapping_node1] = mapping_node2
                property_line_map[mapping_node1] = mapping_line_number
                mapping_rows[mapping_line_number] = mrow.copy()

            else:
                print("Unknown mapping action %s at line %d of mapping file %s" % (mapping_label,
                                                                                   mapping_line_number,
                                                                                   repr(str(mapping_kgtk_file))),
                      file=error_file, flush=True)
                mapping_errors += 1
                continue
                

        # Close the mapping file.
        mkr.close()
        if rmkw is not None:
            rmkw.close()

        if mapping_errors > 0:
            raise KGTKException("%d errors detected in the mapping file %s" % (mapping_errors, repr(str(mapping_kgtk_file))))

        if len(item_map) == 0 and len(property_map) == 0:
            raise KGTKException("Nothing read from the mapping file %s" % repr(str(mapping_kgtk_file)))

        if verbose:
            print("%d mapping lines, %d excluded for confidence, %d excluded for idempotency." % (mapping_line_number,
                                                                                                  mapping_confidence_exclusions,
                                                                                                  mapping_idempotent_exclusions),
                  file=error_file, flush=True)
            print("%d item mapping rules." % len(item_map), file=error_file, flush=True)
            print("%d property mapping rules." % len(property_map), file=error_file, flush=True)

        if verbose:
            print("Opening the input file %s." % repr(str(input_kgtk_file)), file=error_file, flush=True)
        ikr:  KgtkReader = KgtkReader.open(input_kgtk_file,
                                           options=input_reader_options,
                                           value_options = value_options,
                                           error_file=error_file,
                                           verbose=verbose,
                                           very_verbose=very_verbose,
        )
        trouble = False
        input_node1_idx: int = ikr.get_node1_column_index(node1_column_name)
        input_label_idx: int = ikr.get_label_column_index(label_column_name)
        input_node2_idx: int = ikr.get_node2_column_index(node2_column_name)
        if input_node1_idx < 0 and mapping_rule_mode in ["normal", "same-as-item"]:
            trouble = True
            print("Error: Cannot find the input file node1 column.", file=error_file, flush=True)
        if input_label_idx < 0 and mapping_rule_mode in ["normal", "same-as-property"]:
            trouble = True
            print("Error: Cannot find the input file label column.", file=error_file, flush=True)
        if input_node2_idx < 0 and mapping_rule_mode in ["normal", "same-as-item"]:
            trouble = True
            print("Error: Cannot find the input file node2 column.", file=error_file, flush=True)
        if trouble:
            # Clean up:                                                                                                                                               
            ikr.close()
            raise KGTKException("Missing columns in the input file.")

        okw: KgtkWriter = KgtkWriter.open(ikr.column_names,
                                          output_kgtk_file,
                                          mode=KgtkWriter.Mode[ikr.mode.name],
                                          use_mgzip=input_reader_options.use_mgzip, # Hack!
                                          mgzip_threads=input_reader_options.mgzip_threads, # Hack!
                                          error_file=error_file,
                                          verbose=verbose,
                                          very_verbose=very_verbose)

        uekw: typing.Optional[KgtkWriter] = None
        if unmodified_edges_kgtk_file is not None:
            if verbose:
                print("Opening the unmodified edges file %s." % repr(str(unmodified_edges_kgtk_file)), file=error_file, flush=True)
            uekw = KgtkWriter.open(ikr.column_names,
                                   unmodified_edges_kgtk_file,
                                   mode=KgtkWriter.Mode[ikr.mode.name],
                                   use_mgzip=input_reader_options.use_mgzip, # Hack!
                                   mgzip_threads=input_reader_options.mgzip_threads, # Hack!
                                   error_file=error_file,
                                   verbose=verbose,
                                   very_verbose=very_verbose)

        amkw: typing.Optional[KgtkWriter] = None
        if activated_mapping_kgtk_file is not None:
            if verbose:
                print("Opening the activated mapping edges file %s." % repr(str(activated_mapping_kgtk_file)), file=error_file, flush=True)
            amkw = KgtkWriter.open(mkr.column_names,
                                   activated_mapping_kgtk_file,
                                   mode=KgtkWriter.Mode[mkr.mode.name],
                                   use_mgzip=input_reader_options.use_mgzip, # Hack!
                                   mgzip_threads=input_reader_options.mgzip_threads, # Hack!
                                   error_file=error_file,
                                   verbose=verbose,
                                   very_verbose=very_verbose)

        # Process each row of the input file.
        if verbose:
            print("Processing the input file.", file=error_file, flush=True)
        input_count: int = 0
        modified_edge_count: int = 0
        unmodified_edge_count: int = 0
        row: typing.List[str]
        for row in ikr:
            input_count +=1
            newrow: typing.List[str] = row.copy()

            modified_node1: bool = False
            modified_node2: bool = False
            modified_label: bool = False

            if mapping_rule_mode in ["normal", "same-as-item"]:
                input_node1: str = row[input_node1_idx]
                if input_node1 in item_map:
                    newrow[input_node1_idx] = item_map[input_node1]
                    modified_node1 = True
                    if amkw is not None:
                        mapping_line_number = item_line_map[input_node1]
                        if mapping_line_number not in activated_mapping_rows:
                            activated_mapping_rows[mapping_line_number] = mapping_rows[mapping_line_number]
                        
                input_node2: str = row[input_node2_idx]
                if input_node2 in item_map:
                    newrow[input_node2_idx] = item_map[input_node2]
                    modified_node2 = True
                    if amkw is not None:
                        mapping_line_number = item_line_map[input_node2]
                        if mapping_line_number not in activated_mapping_rows:
                            activated_mapping_rows[mapping_line_number] = mapping_rows[mapping_line_number]

            if mapping_rule_mode in ["normal", "same-as-property"]:
                input_label: str = row[input_label_idx]
                if input_label in property_map:
                    newrow[input_label_idx] = property_map[input_label]
                    modified_label = True
                    if amkw is not None:
                        mapping_line_number = property_line_map[input_label]
                        if mapping_line_number not in activated_mapping_rows:
                            activated_mapping_rows[mapping_line_number] = mapping_rows[mapping_line_number]

            modified: bool
            if modified_pattern == "node1|label|node2":
                modified = modified_node1 or modified_label or modified_node2
            elif modified_pattern == "node1|label":
                modified = modified_node1 or modified_label
            elif modified_pattern == "node1|node2":
                modified = modified_node1 or modified_node2
            elif modified_pattern == "label|node2":
                modified = modified_label or modified_node2
            elif modified_pattern == "node1":
                modified = modified_node1
            elif modified_pattern == "label":
                modified = modified_label
            elif modified_pattern == "node2":
                modified = modified_node2
            elif modified_pattern == "node1&label&node2":
                modified = modified_node1 and modified_label and modified_node2
            elif modified_pattern == "node1&label":
                modified = modified_node1 and modified_label
            elif modified_pattern == "node1&node2":
                modified = modified_node1 and modified_node2
            elif modified_pattern == "label&node2":
                modified = modified_label and modified_node2
            else:
                raise KGTKException("Unrecognized modification test pattern %s" % repr(modified_pattern))                

            if modified:
                modified_edge_count += 1
                okw.write(newrow)
            else:
                unmodified_edge_count += 1
                if uekw is not None:
                    uekw.write(row)
                if not split_output_mode:
                    okw.write(row)
                        
        # Done!
        ikr.close()
        okw.close()

        if verbose:
            print("%d edges read. %d modified, %d unmodified." % (input_count, modified_edge_count, unmodified_edge_count), file=error_file, flush=True)

        if uekw is not None:
            uekw.close()

        if amkw is not None:
            activated_count: int = 0
            for mapping_line_number in sorted(activated_mapping_rows.keys()):
                amkw.write(activated_mapping_rows[mapping_line_number])
                activated_count += 1
            amkw.close()

            if verbose:
                print("%d activated mapping edges" % activated_count, file=error_file, flush=True)

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




"""
Import ConceptNet into KGTK.

"""

import sys
from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Import ConceptNet into KGTK.' 
    }

def add_arguments(parser: KGTKArgumentParser):
    """
    Parse arguments
    Args:
            parser (argparse.ArgumentParser)
    """
    parser.add_input_file(positional=True)
    parser.add_argument('--english_only', action="store_true", help="Only english conceptnet?")
    parser.add_output_file()
    parser.add_output_file(who="A KGTK output file that will contain only the weights.",
                           dest="weights_file",
                           options=["--weights-file"],
                           metavar="WEIGHTS_FILE",
                           optional=True)

def run(input_file: KGTKFiles, english_only, output_file: KGTKFiles, weights_file: KGTKFiles):

    # import modules locally
    import sys # type: ignore
    from kgtk.exceptions import kgtk_exception_auto_handler
    import csv
    import json
    import re
    from pathlib import Path
    from kgtk.kgtkformat import KgtkFormat
    from kgtk.io.kgtkwriter import KgtkWriter

    def make_node_label(node):
        return KgtkFormat.stringify(node.strip().split('/')[3].replace('_', ' '))

    def split_camel_case(name):
        splitted = re.sub('([A-Z][a-z]+)', r' \1', re.sub('([A-Z]+)', r' \1', name)).split()
        return ' '.join(splitted).lower()

    def make_rel_label(rel):
        return KgtkFormat.stringify(split_camel_case(rel.split('/')[-1]))

    def make_weight_edge(row):

        node1='%s-%s-%s-0000' % (row[2], row[1], row[3])
        rel='weight'
        node2=str(json.loads(row[-1])['weight'])
        return [node1, rel, node2]

    def row_to_edge(row, cols):

        edge={}
        edge['node1']=row[2]
        edge['relation']=row[1]
        edge['node2']=row[3]
        edge['node1;label']=make_node_label(row[2])
        edge['node2;label']=make_node_label(row[3])
        edge['relation;label']=make_rel_label(row[1])
        edge['relation;dimension']=''

        metadata=json.loads(row[4])
        edge['source']=KgtkFormat.stringify('CN')
        if 'surfaceText' in metadata.keys():
            edge['sentence']=KgtkFormat.stringify(metadata['surfaceText'].replace('\\', ''))
        else:
            edge['sentence']=''
        
        edge_list=[edge[col] for col in cols]
        return edge_list

    try:
        filename: Path = KGTKArgumentParser.get_input_file(input_file)

        in_columns=['assertion','rel','subj','obj','metadata']
        out_columns=['node1', 'relation', 'node2', 'node1;label', 'node2;label','relation;label', 'relation;dimension', 'source', 'sentence']

        output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)
        ew: KgtkWriter = KgtkWriter.open(out_columns,
                                         output_kgtk_file,
                                         #mode=input_kr.mode,
                                         require_all_columns=False,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=True,
                                         gzip_in_parallel=False,
                                         #verbose=self.verbose,
                                         #very_verbose=self.very_verbose
                                         )

        if weights_file:
            info_kgtk_file: Path = KGTKArgumentParser.get_output_file(weights_file)
            ew_aux: KgtkWriter = KgtkWriter.open(out_columns[:3],
                                             info_kgtk_file,
                                             #mode=input_kr.mode,
                                             require_all_columns=False,
                                             prohibit_extra_columns=True,
                                             fill_missing_columns=True,
                                             gzip_in_parallel=False,
                                             #verbose=self.verbose,
                                             #very_verbose=self.very_verbose
                                             )

        with open(filename, 'r') as f:
            reader = csv.reader(f, delimiter='\t', quotechar='"')
            for row in reader:
                if not english_only or (row[2].startswith('/c/en/') and row[3].startswith('/c/en/')):
                    ew.write(row_to_edge(row, out_columns))
                    if weights_file and 'weight' in json.loads(row[-1]).keys():
                        ew_aux.write(make_weight_edge(row))

        # Clean up
        ew.close()
        if weights_file:
            ew_aux.close()

    except Exception as e:
        kgtk_exception_auto_handler(e)



"""
Normalize a KGTK node file by creating an edge file with a row for each column value.
"""
from argparse import Namespace, SUPPRESS
from pathlib import Path
import sys
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles
from kgtk.kgtkformat import KgtkFormat
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalue import KgtkValue
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

def parser():
    return {
        'help': 'Normalize a KGTK node file into a KGTK edge file.',
        'description': 'Normalize a KGTK node file into a KGTK edge file with a row for each column value in the input file.'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """

    _expert: bool = parsed_shared_args._expert
    _mode: str = parsed_shared_args._mode

    parser.add_input_file(positional=True)
    parser.add_output_file()

    parser.add_argument('-c', "--columns", action="store", type=str, dest="columns", nargs='+',
                        help="Columns to remove as a space-separated list. (default=all columns except id)")

    parser.add_argument(      "--labels", action="store", type=str, dest="labels", nargs='+',
                        help="Label names to use as a space-separated list. (default=column names)")

    parser.add_argument(      "--id-column", action="store", type=str, dest="id_column_name",
                        help="The name of the ID column. (default=id or alias)")

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode.NONE if _mode == "NONE" else KgtkReaderMode.NODE,
                                    expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,

        columns: typing.Optional[typing.List[str]] = None,
        labels: typing.Optional[typing.List[str]] = None,
        id_column_name: typing.Optional[str] = None,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    import os

    from kgtk.exceptions import kgtk_exception_auto_handler, KGTKException

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_kgtk_file), file=error_file)
        print("--output-file=%s" % str(output_kgtk_file), file=error_file)

        if columns is not None:
            print("--columns=%s" % " ".join(columns), file=error_file)
        if labels is not None:
            print("--labels=%s" % " ".join(labels), file=error_file)
        if id_column_name is not None:
            print("--id-column=%s" % id_column_name, file=error_file)

        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    if verbose:
        print("Starting normalize_nodes pid=%d" % (os.getpid()), file=error_file, flush=True)

    label_map: typing.MutableMapping[str, str] = dict()
    if labels is not None and len(labels) > 0:
        if columns is None:
            raise KGTKException("--columns must be supplied when --labels is used.")
        if len(columns) != len(labels):
            raise KGTKException("%d columns were supplied, but %d labels." % (len(columns), len(labels)))
        idx: int
        label: str
        for idx, label in enumerate(labels):
            label_map[columns[idx]] = label

    try:
        if verbose:
            print("Opening the input file: %s" % str(input_kgtk_file), file=error_file, flush=True)
        kr: KgtkReader = KgtkReader.open(input_kgtk_file,
                                         options=reader_options,
                                         value_options = value_options,
                                         error_file=error_file,
                                         verbose=verbose,
                                         very_verbose=very_verbose,
        )

        id_column_idx: int = kr.get_id_column_index(id_column_name)
        if id_column_idx < 0:
            raise KGTKException("Unknown ID column %s" % repr(id_column_name))

        output_column_names: typing.List[str] = [ KgtkFormat.NODE1, KgtkFormat.LABEL, KgtkFormat.NODE2 ]

        if verbose:
            print("Opening the output file: %s" % str(output_kgtk_file), file=error_file, flush=True)
        kw: KgtkWriter = KgtkWriter.open(output_column_names,
                                         output_kgtk_file,
                                         mode=KgtkWriter.Mode.EDGE,
                                         verbose=verbose,
                                         very_verbose=very_verbose)

        input_line_count: int = 0
        output_line_count: int = 0
        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            node1_value: str = row[id_column_idx]

            column_idx: int
            column_name: str
            for column_idx, column_name in enumerate(kr.column_names):
                if column_idx == id_column_idx:
                    continue
                if columns is not None and column_name not in columns:
                    continue

                label_value: str = label_map.get(column_name, column_name)

                new_value: str = row[column_idx]
                if len(new_value) == 0:
                    continue # ignore empty values.

                # The column value might contain a KGTK list.  Since node2 isn't supposed
                # to contain lists, we'll split it.
                node2_value: str
                for node2_value in KgtkValue.split_list(new_value):
                    if len(node2_value) == 0:
                        continue # node2 shouldn't contain empty values

                    output_row: typing.List[str] = [ node1_value , label_value, node2_value ]
                    kw.write(output_row)
                    output_line_count += 1

        if verbose:
            print("Read %d node rows, wrote %d edge rows." % (input_line_count, output_line_count), file=error_file, flush=True)

        kw.close()

        return 0

    except Exception as e:
        kgtk_exception_auto_handler(e)
        return 1



import unittest
import os
from kgtk.generator import JsonGenerator
from pathlib import Path


class TestJSONGeneration(unittest.TestCase):
    def test_dates_generation(self):
        dates_tsv_file = Path('data/dates.tsv')
        wikidata_property_file = 'data/wikidata_properties.tsv'
        generator = JsonGenerator(prop_file=Path(wikidata_property_file), label_set='label', alias_set='aliase',
                                  description_set='description', warning=True, n=100,
                                  log_path="data/date_warning.log",
                                  has_rank=False,
                                  prop_declaration=False,
                                  output_prefix="data/dates_tmp",
                                  input_file=Path(dates_tsv_file),
                                  property_declaration_label="property_type",
                                  error_action='log')
        generator.process()

        f1 = open('data/dates0.jsonl')
        f2 = open('data/dates_tmp0.jsonl')
        self.assertEqual(f1.readlines(), f2.readlines())

        f1.close()
        f2.close()
        self.assertEqual(os.stat("data/date_warning.log").st_size, 0)
        p = Path("data/date_warning.log")
        p.unlink()
        p = Path('data/dates_tmp0.jsonl')
        p.unlink()

    def test_property_json_generation(self):
        property_tsv_file = Path('data/P10.tsv')
        wikidata_property_file = 'data/wikidata_properties.tsv'
        generator = JsonGenerator(prop_file=Path(wikidata_property_file), label_set='label', alias_set='aliases',
                                  description_set='descriptions', warning=True, n=1000,
                                  log_path="data/P10_warning.log",
                                  has_rank=False,
                                  prop_declaration=False,
                                  output_prefix="data/P10_tmp",
                                  property_declaration_label="property_type",
                                  input_file=Path(property_tsv_file), error_action='log')
        generator.process()

        f1 = open('data/P100.jsonl')
        f2 = open('data/P10_tmp0.jsonl')
        self.assertEqual(f1.readlines(), f2.readlines())
        f1.close()
        f2.close()
        self.assertEqual(os.stat("data/P10_warning.log").st_size, 0)
        p = Path("data/P10_warning.log")
        p.unlink()
        p = Path('data/P10_tmp0.jsonl')
        p.unlink()

    def test_qnode_json_generation(self):
        qnode_tsv_file = Path('data/Q57160439.tsv')
        wikidata_property_file = 'data/wikidata_properties.tsv'
        generator = JsonGenerator(prop_file=Path(wikidata_property_file), label_set='label', alias_set='aliases',
                                  description_set='descriptions', warning=True, n=1000,
                                  log_path="data/Q57160439_warning.log",
                                  prop_declaration=False,
                                  has_rank=False,
                                  output_prefix="data/Q57160439_tmp",
                                  property_declaration_label="property_type",
                                  input_file=Path(qnode_tsv_file), error_action='log')
        generator.process()

        f1 = open('data/Q571604390.jsonl')
        f2 = open('data/Q57160439_tmp0.jsonl')
        self.assertEqual(f1.readlines(), f2.readlines())
        f1.close()
        f2.close()
        self.assertEqual(os.stat("data/Q57160439_warning.log").st_size, 0)
        p = Path("data/Q57160439_warning.log")
        p.unlink()
        p = Path('data/Q57160439_tmp0.jsonl')
        p.unlink()

    def test_ranked_kgtk_generation(self):
        ranked_tsv_file = Path('data/ranked_example.tsv')
        wikidata_property_file = 'data/wikidata_properties.tsv'
        generator = JsonGenerator(prop_file=Path(wikidata_property_file), label_set='label', alias_set='alias',
                                  description_set='description', warning=True, n=1000,
                                  log_path="data/ranked_warning.log",
                                  prop_declaration=False,
                                  has_rank=False,
                                  output_prefix="data/ranked_tmp",
                                  property_declaration_label="property_type",
                                  input_file=Path(ranked_tsv_file),
                                  error_action='log')
        generator.process()

        f1 = open('data/ranked0.jsonl')
        f2 = open('data/ranked_tmp0.jsonl')
        self.assertEqual(f1.readlines(), f2.readlines())
        f1.close()
        f2.close()
        p = Path('data/ranked_tmp0.jsonl')
        p.unlink()



import re
import json
import gzip
from pathlib import Path
import rfc3986
import sys
import typing
from typing import List
from kgtk.knowledge_graph.schema import KGSchema
from kgtk.wikidata import wiki_namespaces
from kgtk.exceptions import KGTKException
from kgtk.wikidata.entity import WDItem, WDProperty
from kgtk.io.kgtkreader import KgtkReader
from kgtk.knowledge_graph.document import Document

from kgtk.wikidata.value import (
    Precision,
    Item,
    StringValue,
    TimeValue,
    QuantityValue,
    MonolingualText,
    GlobeCoordinate,
    ExternalIdentifier,
    URLValue
)
from kgtk.knowledge_graph.node import LiteralType

BAD_CHARS = [":", "&", ",", " ",
             "(", ")", "\'", '\"', "/", "\\", "[", "]", ";", "|"]


class Generator:
    def __init__(self, **kwargs):
        label_set = kwargs.pop("label_set")
        description_set = kwargs.pop("description_set")
        alias_set = kwargs.pop("alias_set")
        n = int(kwargs.pop("n"))
        warning = kwargs.pop("warning")
        self.log_path = kwargs.pop("log_path")
        self.prop_file: typing.Optional[Path] = kwargs.pop("prop_file")
        self.read_num_of_lines = 0
        # set sets
        self.set_sets(label_set, description_set, alias_set)
        # column name order_map
        self.order_map = {}
        self.n = n
        self.yyyy_mm_dd_pattern = re.compile(
            r"[12]\d{3}-(0[1-9]|1[0-2])-(0[1-9]|[12]\d|3[01])")
        self.yyyy_pattern = re.compile(r"[12]\d{3}")
        self.quantity_pattern = re.compile(
            r"([\+|\-]?[0-9]+\.?[0-9]*[e|E]?[\-]?[0-9]*)(?:\[([\+|\
            -]?[0-9]+\.?[0-9]*),([\+|\-]?[0-9]+\.?[0-9]*)\])?([U|Q](?:.*))?")
        self.warning = warning
        if self.warning:
            if self.log_path == '-':
                self.warn_log = sys.stderr
            else:
                self.warn_log = open(self.log_path, "w")
        else:
            self.warn_log = sys.stderr
        self.to_append_statement_id = None
        self.corrupted_statement_id = None
        self.to_append_statement = None  # for Json generator
        self.wiki_import_prop_types = {"wikipedia_sitelink", "language"}
        self.datatype_mapping = {
            # nomenclature from https://w.wiki/Tfn
            "item": Item,
            "WikibaseItem": Item,
            "wikibase-item": Item,

            "time": TimeValue,
            "Time": TimeValue,

            "globe-coordinate": GlobeCoordinate,
            "GlobeCoordinate": GlobeCoordinate,

            "quantity": QuantityValue,
            "Quantity": QuantityValue,

            "monolingualtext": MonolingualText,
            "Monolingualtext": MonolingualText,

            "string": StringValue,
            "String": StringValue,

            "external-identifier": ExternalIdentifier,
            "ExternalId": ExternalIdentifier,
            "external-id": ExternalIdentifier,

            "url": StringValue,  # TODO bug potentially in rdflib
            "Url": StringValue,

            "property": WDProperty,
            "WikibaseProperty": WDProperty,
            "wikibase-property": WDProperty
        }

    def serialize(self):
        raise NotImplemented

    def finalize(self):
        if self.warning and self.log_path != "-":
            self.warn_log.close()
        self.serialize()

    def set_sets(self, label_set: str, description_set: str, alias_set: str):
        self.label_set, self.alias_set, self.description_set = set(label_set.split(",")), set(
            alias_set.split(",")), set(description_set.split(","))

    @staticmethod
    def process_text_string(string: str) -> List[str]:
        """
        Language detection is removed from triple generation. The user is responsible for detect the language
        """
        if len(string) == 0:
            return ["", "en"]
        if "@" in string:
            res = string.split("@")
            text_string = "@".join(res[:-1]).replace('"', "").replace("'", "")
            lang = res[-1].replace('"', '').replace("'", "")
        else:
            text_string = string.replace('"', "").replace("'", "")
            lang = "en"
        return [text_string, lang]

    @staticmethod
    def is_invalid_decimal_string(num_string) -> bool:
        """
        if a decimal string too small, return True TODO
        """
        if num_string is None:
            return False
        else:
            if abs(float(num_string)) < 0.0001 and float(num_string) != 0:
                return True
            return False

    @staticmethod
    def is_valid_uri_with_scheme_and_host(uri: str) -> bool:
        """
        https://github.com/python-hyper/rfc3986/issues/30#issuecomment-461661883
        """
        try:
            uri = rfc3986.URIReference.from_string(uri)
            rfc3986.validators.Validator().require_presence_of(
                "scheme", "host").check_validity_of("scheme", "host").validate(uri)
            return True
        except Exception:
            return False

    @staticmethod
    def clean_number_string(num: typing.Optional[str]) -> typing.Optional[str]:
        from numpy import format_float_positional
        if num is None:
            return None
        else:
            return format_float_positional(float(num), trim="-")

    @staticmethod
    def replace_illegal_string(s: str) -> str:
        """
        this function serves as the last gate of keeping illegal characters outside of entity creation.
        """
        for char in BAD_CHARS:
            s = s.replace(char, "_")
        return s


class TripleGenerator(Generator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        prop_declaration = kwargs.pop("prop_declaration")
        dest_fp = kwargs.pop("dest_fp")
        truthy = kwargs.pop("truthy")
        use_id = kwargs.pop("use_id")
        prefix_path = kwargs.pop("prefix_path")
        self.error_action = kwargs.pop('error_action')
        self.set_prefix(prefix_path)
        self.prop_declaration = prop_declaration
        self.set_properties(self.prop_file)
        if str(dest_fp).endswith('.gz'):
            self.fp = gzip.open(dest_fp, 'wt')
        else:
            self.fp = open(dest_fp, 'w')
        self.truthy = truthy
        self.reset_etk_doc()
        self.serialize_prefix()
        self.use_id = use_id
        self.node1_idx = -1
        self.node2_idx = -1
        self.id_idx = -1
        self.label_idx = -1
        self.kr = None
        self.initialize(kwargs.pop('input_file'))

    def initialize(self, input_file):
        self.input_file = input_file
        self.kr: KgtkReader = KgtkReader.open(self.input_file)
        self.node1_idx = self.kr.get_node1_column_index()
        self.node2_idx = self.kr.get_node2_column_index()
        self.label_idx = self.kr.get_label_column_index()
        self.id_idx = self.kr.get_id_column_index()
        if self.node1_idx == -1:
            raise KGTKException("'node1' column not found")

        if self.node2_idx == -1:
            raise KGTKException("'node2' column not found")

        if self.label_idx == -1:
            raise KGTKException("'label' column not found")

        if self.id_idx == -1:
            raise KGTKException("'id' column not found")

    def process(self):
        input_row_count: int = 2

        if self.prop_declaration:
            for row in self.kr:
                self.read_prop_declaration(row)
            self.kr.close()
            self.kr: KgtkReader = KgtkReader.open(self.input_file)
        for row in self.kr:
            self.entry_point(input_row_count, row)
            input_row_count += 1

        self.finalize()

    def set_prefix(self, prefix_path: str):
        self.prefix_dict = {}
        if prefix_path != "NONE":
            with open(prefix_path, "r") as fp:
                for line_num, edge in enumerate(fp):
                    edge_list = edge.strip("\r\n").split("\t")
                    if line_num == 0:
                        node1_index, node2_index = edge_list.index("node1"), edge_list.index("node2")
                    else:
                        prefix, expand = edge_list[node1_index], edge_list[node2_index]
                        self.prefix_dict[prefix] = expand

    def read_prop_declaration(self, row: List[str]):
        node1, node2, prop, e_id = row[self.node1_idx], row[self.node2_idx], row[self.label_idx], row[self.id_idx]
        if prop == "data_type" or prop == "datatype":
            self.prop_types[node1] = self.datatype_mapping[node2.strip()]

    def set_properties(self, prop_file: typing.Optional[Path]):
        self.prop_types = dict()
        if prop_file is None:
            return

        if str(prop_file).endswith(".gz"):
            fp = gzip.open(prop_file, 'rt')
        else:
            fp = open(prop_file, "r")

        node1_idx = -1
        node2_idx = -1

        for line in fp:
            vals = line.split('\t')
            vals = [v.strip() for v in vals]
            if 'node1' in vals and 'node2' in vals:
                node1_idx = vals.index('node1')
                node2_idx = vals.index('node2')

            else:
                node1 = vals[node1_idx]
                node2 = vals[node2_idx]
                try:
                    self.prop_types[node1] = self.datatype_mapping[node2.strip()]
                except Exception:
                    self.prop_types[node1] = StringValue
                    if self.error_action == 'ignore':
                        pass
                    elif self.error_action == 'log':
                        self.warn_log.write(
                            "DataType {} of node {} is not supported. "
                            "{}'s DataType has been defaulted to StringValue.\n".format(node2, node1, node1)
                        )
                    elif self.error_action == 'raise':
                        raise KGTKException("DataType {} of node {} is not supported.".format(node2, node1))
                    else:
                        raise KGTKException("Unknown error_action {} processing unsupported data type {} of node {}."
                                            .format(self.error_action, node2, node1))

    def _node_2_entity(self, node: str):
        '''
        A node can be Qxxx or Pxxx, return the proper entity.
        '''
        if node in self.prop_types:
            entity = WDProperty(node, self.prop_types[node])
        else:
            entity = WDItem(TripleGenerator.replace_illegal_string(node))
        return entity

    def reset_etk_doc(self, doc_id: str = "http://isi.edu/default-ns/projects"):
        """
        reset the doc object and return it. Called at initialization and after outputting triples.
        """
        kg_schema = KGSchema()
        kg_schema.add_schema("@prefix : <http://isi.edu/> .", "ttl")
        self.doc = Document({}, kg_schema, doc_id=doc_id)
        for k, v in wiki_namespaces.items():
            if k not in self.prefix_dict:
                self.doc.kg.bind(k, v)
        for k, v in self.prefix_dict.items():
            self.doc.kg.bind(k, v)

    def serialize(self):
        """
        Serialize the triples. Used a hack to avoid serializing the prefix again.
        """
        self.fp.write("\n\n".join(
            self.doc.kg.serialize("ttl").split("\n\n")[1:]))
        self.fp.flush()
        self.reset()

    def serialize_prefix(self):
        """
        This function should be called only once after the doc object is initialized.
        In order to serialize the prefix at the very begining it has to be printed per the change of rdflib 4.2.2->5.0.0
        Relevent issue: https://github.com/RDFLib/rdflib/issues/965
        """
        for k, v in wiki_namespaces.items():
            if k not in self.prefix_dict:
                self.fp.write("@prefix " + k + ": <" + v + "> .\n")

        # Add the following additional prefixes.  Other prefixes (from the prefixes known to
        # rdflib) might be generated.
        #
        # TODO: we need a principled solution to the problem of emitting all
        # required prefixes (and preferably, only required prefixes).
        if "rdfs" not in self.prefix_dict:
            self.fp.write("@prefix " + "rdfs" + ": <" + "http://www.w3.org/2000/01/rdf-schema#" + "> .\n")
        if "xsd" not in self.prefix_dict:
            self.fp.write("@prefix " + "xsd" + ": <" + "http://www.w3.org/2001/XMLSchema#" + "> .\n")

        for k, v in self.prefix_dict.items():
            self.fp.write("@prefix " + k + ": <" + self.prefix_dict[k] + "> .\n")

        self.fp.write("\n")
        self.fp.flush()
        self.reset()

    def reset(self):
        self.to_append_statement_id = None
        self.to_append_statement = None
        self.read_num_of_lines = 0
        self.reset_etk_doc()

    def generate_label_triple(self, node1: str, node2: str) -> bool:
        entity = self._node_2_entity(node1)
        text_string, lang = TripleGenerator.process_text_string(node2)
        entity.add_label(text_string, lang=lang)
        self.doc.kg.add_subject(entity)
        return True

    def generate_description_triple(self, node1: str, node2: str) -> bool:
        entity = self._node_2_entity(node1)
        text_string, lang = TripleGenerator.process_text_string(node2)
        entity.add_description(text_string, lang=lang)
        self.doc.kg.add_subject(entity)
        return True

    def generate_alias_triple(self, node1: str, node2: str) -> bool:
        entity = self._node_2_entity(node1)
        text_string, lang = TripleGenerator.process_text_string(node2)
        entity.add_alias(text_string, lang=lang)
        self.doc.kg.add_subject(entity)
        return True

    def generate_prop_declaration_triple(self, node1: str, node2: str) -> bool:
        # update the known prop_types
        if node1 in self.prop_types:
            if not self.prop_declaration:
                if self.error_action == 'ignore':
                    pass
                elif self.error_action == 'log':
                    self.warn_log.write("IMPORTANT: Duplicated property definition of {} found!."
                                        "Using data type: {} for property {}".format(node1, self.prop_types[node1],
                                                                                     node1))
                elif self.error_action == 'raise':
                    raise KGTKException("Duplicated property definition of {} found!".format(node1))
                else:
                    raise KGTKException("Unknown error_action {} processing duplicated property definition of {}."
                                        .format(self.error_action, node1))
        else:
            self.prop_types[node1] = node2

        prop = WDProperty(node1, self.datatype_mapping.get(node2, StringValue))
        self.doc.kg.add_subject(prop)
        return True

    def generate_normal_triple(
            self, node1: str, property: str, node2: str, is_qualifier_edge: bool, e_id: str, line_number: int) -> bool:
        if self.use_id:
            e_id = TripleGenerator.replace_illegal_string(e_id)
        entity = self._node_2_entity(node1)
        edge_type = self.prop_types[property]
        if edge_type == Item:
            object = WDItem(TripleGenerator.replace_illegal_string(node2))
        elif edge_type == WDProperty:
            object = WDProperty(TripleGenerator.replace_illegal_string(node2), self.prop_types[node2])

        elif edge_type == TimeValue:
            if self.yyyy_mm_dd_pattern.match(node2):
                try:
                    dateTimeString = node2
                    object = TimeValue(
                        value=dateTimeString,  # TODO
                        calendar=Item("Q1985727"),
                        precision=Precision.year,
                        time_zone=0,
                    )
                except Exception:
                    return False
            elif self.yyyy_pattern.match(node2):
                try:
                    dateTimeString = node2 + "-01-01"
                    object = TimeValue(
                        value=dateTimeString,  # TODO
                        calendar=Item("Q1985727"),
                        precision=Precision.year,
                        time_zone=0,
                    )
                except Exception:
                    return False
            else:
                try:
                    # TODO, in future, the two cases above will be dropped in principle to comply with the iso format
                    # now it is iso format
                    assert (node2[0] == "^")
                    node2 = node2[1:]  # remove ^
                    if node2.startswith("+"):
                        node2 = node2[1:]
                    dateTimeString, precision = node2.split("/")
                    dateTimeString = dateTimeString[:-1]  # remove Z
                    object = TimeValue(
                        value=dateTimeString,
                        calendar=Item("Q1985727"),
                        precision=precision,
                        time_zone=0,
                    )
                except Exception as e:
                    print(e)
                    return False

        elif edge_type == GlobeCoordinate:
            latitude_str: str
            longitude_str: str
            latitude_str, longitude_str = node2[1:].split("/")
            latitude: float = float(latitude_str)
            longitude: float = float(longitude_str)
            object = GlobeCoordinate(
                latitude, longitude, 0.0001, globe=Item("Q2"))  # earth

        elif edge_type == QuantityValue:
            try:
                res = self.quantity_pattern.match(node2)
                if res is None:
                    if self.warning:
                        self.warn_log.write("Node2 [{}] at line [{}] is not a legal quantity. Skipping it.\n".format(
                            node2, line_number))
                    return False
                res = res.groups()

            except Exception:
                if self.error_action == 'ignore':
                    pass
                elif self.error_action == 'log':
                    self.warn_log.write(
                        "Node2 [{}] at line [{}] is not a legal quantity.\n".format(
                            node2, line_number)
                    )
                elif self.error_action == 'raise':
                    raise KGTKException(
                        "Node2 [{}] at line [{}] is not a legal quantity.\n".format(
                            node2, line_number)
                    )
                else:
                    raise KGTKException("Unknown error_action {} processing illegal quntity in node2 [{}] at line [{}]."
                                        .format(self.error_action, node2, line_number))

            amount, lower_bound, upper_bound, unit = res

            amount: typing.Optional[str] = TripleGenerator.clean_number_string(amount)
            num_type = self.xsd_number_type(amount)  # Error! xsd_number_type expectes a float or int!

            lower_bound = TripleGenerator.clean_number_string(lower_bound)
            upper_bound = TripleGenerator.clean_number_string(upper_bound)
            if unit is not None:
                if upper_bound is not None and lower_bound is not None:
                    object = QuantityValue(amount, unit=Item(
                        unit), upper_bound=upper_bound, lower_bound=lower_bound, type=num_type)
                else:
                    object = QuantityValue(amount, unit=Item(unit), type=num_type)
            else:
                if upper_bound is not None and lower_bound is not None:
                    object = QuantityValue(
                        amount, upper_bound=upper_bound, lower_bound=lower_bound, type=num_type)
                else:
                    object = QuantityValue(amount, type=num_type)

        elif edge_type == MonolingualText:
            text_string, lang = TripleGenerator.process_text_string(node2)
            object = MonolingualText(text_string, lang)
        elif edge_type == ExternalIdentifier:
            object = ExternalIdentifier(node2)
        elif edge_type == URLValue:
            if TripleGenerator.is_valid_uri_with_scheme_and_host(node2):
                object = URLValue(node2)
            else:
                return False
        else:
            # treat everything else as stringValue
            object = StringValue(node2)

        if type(object) == WDItem or type(object) == WDProperty:
            self.doc.kg.add_subject(object)

        if is_qualifier_edge:
            # edge: e8 p9 ^2013-01-01T00:00:00Z/11
            # create qualifier edge on previous STATEMENT and return the updated STATEMENT
            if self.to_append_statement is None:
                raise KGTKException("Qualifier edge with no preceeding statement at line %d: (%s, %s, %s)" %
                                    (line_number, repr(node1), repr(property), repr(node2)))
            self.to_append_statement.add_qualifier(property, object)
            self.doc.kg.add_subject(self.to_append_statement)
        else:
            # edge: q1 p8 q2 e8
            # create brand new property edge and replace STATEMENT
            if self.truthy:
                self.to_append_statement = entity.add_truthy_statement(
                    property, object, statement_id=e_id) if self.use_id else entity.add_truthy_statement(property,
                                                                                                         object)
            else:
                self.to_append_statement = entity.add_statement(
                    property, object, statement_id=e_id) if self.use_id else entity.add_statement(property, object)
            self.doc.kg.add_subject(entity)
        return True

    def entry_point(self, line_number: int, row: List[str]):
        """
        generates a list of two, the first element is the determination of the edge type using corresponding edge type
        the second element is a bool indicating whether this is a valid property edge or qualifier edge.
        Call corresponding downstream functions
        """

        success = True
        node1, node2, prop, e_id = row[self.node1_idx], row[self.node2_idx], row[self.label_idx], row[self.id_idx]
        if line_number == 2:
            # by default a statement edge
            is_qualifier_edge = False
        else:
            if node1 != self.to_append_statement_id and node1 != self.corrupted_statement_id:
                is_qualifier_edge = False
                # also a new statement edge
                if self.read_num_of_lines >= self.n:
                    self.serialize()
            else:
                # qualifier edge or property declaration edge
                is_qualifier_edge = True
                if node1 == self.corrupted_statement_id:
                    if self.warning:
                        self.warn_log.write(
                            "QUALIFIER edge at line [{}] associated of corrupted statement edge of id [{}] "
                            "dropped.\n".format(line_number, self.corrupted_statement_id))
                    return
        if prop in self.label_set:
            success = self.generate_label_triple(node1, node2)
        elif prop in self.description_set:
            success = self.generate_description_triple(node1, node2)
        elif prop in self.alias_set:
            success = self.generate_alias_triple(node1, node2)
        elif prop == "data_type" or prop == 'datatype':
            # special edge of prop declaration
            success = self.generate_prop_declaration_triple(node1, node2)
        else:
            if prop in self.prop_types:
                success = self.generate_normal_triple(
                    node1, prop, node2, is_qualifier_edge, e_id, line_number)
            else:
                if self.error_action == 'ignore':
                    pass
                elif self.error_action == 'log':
                    self.warn_log.write("IMPORTANT: property [{}]'s type is unknown at line [{}].\n".format(
                        prop, line_number))
                elif self.error_action == 'raise':
                    raise KGTKException(
                        "property [{}]'s type is unknown at line [{}].\n".format(
                            prop, line_number)
                    )
                else:
                    raise KGTKException("Unknown error_action {} processing property [{}] with unknown type at line "
                                        "[{}].".format(self.error_action, prop, line_number))

        if (not success):
            if not is_qualifier_edge:
                if self.warning:
                    self.warn_log.write(
                        "CORRUPTED_STATEMENT edge at line: [{}] with edge id [{}].\n".format(
                            line_number, e_id))
                self.corrupted_statement_id = e_id
            else:
                if self.warning:
                    self.warn_log.write(
                        "CORRUPTED_QUALIFIER edge at line: [{}] with edge id [{}].\n".format(
                            line_number, e_id))

        else:
            self.read_num_of_lines += 1
            if not is_qualifier_edge:
                self.to_append_statement_id = e_id

    @staticmethod
    def xsd_number_type(num):
        if isinstance(num, float) and 'e' in str(num).lower():
            return LiteralType.double
        return LiteralType.decimal


class JsonGenerator(Generator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_declaration: bool = kwargs.pop("prop_declaration")
        self.output_prefix: str = kwargs.pop("output_prefix")
        self.has_rank: bool = kwargs.pop("has_rank")
        self.error_action: str = kwargs.pop('error_action')
        self.property_declaration_label: str = kwargs.pop("property_declaration_label") \
            if "property_declaration_label" in kwargs else "data_tyoe"
        self.ignore_property_declarations_in_file: bool = kwargs.pop("ignore_property_declarations_in_file") \
            if "ignore_property_declarations_in_file" in kwargs else True
        self.filter_prop_file: bool = kwargs.pop("filter_prop_file") if "filter_prop_file" in kwargs else True
        self.verbose: bool = kwargs.pop("verbose") if "verbose" in kwargs else False
        self.file_num = 0
        # this data_type mapping is to comply with the SQID UI parsing requirements
        self.datatype_mapping = {
            "item": "wikibase-item",
            "WikibaseItem": "wikibase-item",
            "wikibase-item": "wikibase-item",

            "property": "wikibase-item",
            "WikibaseProperty": "wikibase-item",
            "wikibase-property": "wikibase-item",

            "time": "time",
            "Time": "time",

            "globe-coordinate": "globe-coordinate",
            "GlobeCoordinate": "globe-coordinate",

            "quantity": "quantity",
            "Quantity": "quantity",

            "monolingualtext": "monolingualtext",
            "Monolingualtext": "monolingualtext",

            "string": "string",
            "String": "string",

            "external-identifier": "external-id",
            "ExternalId": "external-id",
            "external-id": "external-id",

            "url": "url",
            "Url": "url"
        }
        self.set_properties(self.prop_file)
        self.set_json_dict()
        self.previous_qnode = None
        self.node1_idx = -1
        self.node2_idx = -1
        self.id_idx = -1
        self.label_idx = -1
        self.kr = None
        self.initialize(kwargs.pop('input_file'))

    def initialize(self, input_file):
        self.input_file = input_file

        if self.verbose:
            print("Opening the input file %s" % repr(str(self.input_file)), file=sys.stderr, flush=True)
        self.kr: KgtkReader = KgtkReader.open(self.input_file)

        self.node1_idx = self.kr.get_node1_column_index()
        if self.node1_idx < 0:
            raise KGTKException("The input file [{}] does not have a node1 column or its alias."
                                .format(str(self.input_file)))
        self.node2_idx = self.kr.get_node2_column_index()
        if self.node2_idx < 0:
            raise KGTKException("The input file [{}] does not have a node2 column or its alias."
                                .format(str(self.input_file)))
        self.label_idx = self.kr.get_label_column_index()
        if self.label_idx < 0:
            raise KGTKException("The input file [{}] does not have a label column or its alias."
                                .format(str(self.input_file)))
        self.id_idx = self.kr.get_id_column_index()
        if self.id_idx < 0:
            raise KGTKException("The input file [{}] does not have an id column or its alias."
                                .format(str(self.input_file)))

        if hasattr(self, 'has_rank') and self.has_rank:
            rank_index = self.kr.get_node1_column_index('rank')
            self.rank_idx = rank_index

    def process(self):
        line_num: int
        row: typing.List[str]
        for line_num, row in enumerate(self.kr):
            self.entry_point(line_num + 1, row)

        if self.verbose:
            print("Closing the input file, %d lines read." % (line_num + 1), file=sys.stderr, flush=True)
        self.kr.close()

        self.finalize()

    def entry_point(self, line_number: int, row: typing.List[str]):
        self.read_num_of_lines += 1

        node1 = row[self.node1_idx]
        node2 = row[self.node2_idx]
        prop = row[self.label_idx]
        e_id = row[self.id_idx]
        if self.has_rank:
            rank = row[self.rank_idx]
        else:
            rank = "normal"  # TODO default rank

        # property declaration
        if prop == self.property_declaration_label:
            if self.prop_declaration:
                self.set_property(node1, node2, line_number)
            elif self.ignore_property_declarations_in_file:
                pass
            else:
                if self.warning:
                    self.warn_log.write(
                        "CORRUPTED_STATEMENT property declaration edge at line: [{}] with edge id [{}].\n".format(
                            line_number, e_id))
            return

        # add qualifier logic
        if line_number == 1:
            self.previous_qnode = node1
            is_qualifier_edge = False
        else:
            if node1 != self.to_append_statement_id and node1 != self.corrupted_statement_id:
                is_qualifier_edge = False
                # partial serialization
                if self.n > 0 and self.read_num_of_lines >= self.n:
                    if self.previous_qnode != node1:
                        self.serialize()
                    # update previous qnode to avoid breaking continuous same-qnode statements into two jsonl files
                self.previous_qnode = node1
            else:
                is_qualifier_edge = True
                if node1 == self.corrupted_statement_id:
                    if self.warning:
                        self.warn_log.write(
                            "QUALIFIER edge at line [{}] associated with corrupted statement edge of id [{}] dropped.\n"
                            .format(line_number, self.corrupted_statement_id))

        # update info_json_dict
        if not is_qualifier_edge:
            if node1 in self.prop_types:
                success = self.update_misc_json_dict_info(node1, line_number, self.prop_types[node1])
            else:
                success = self.update_misc_json_dict_info(node1, line_number, None)
            assert (success)

        if prop in self.prop_types:
            success = self.update_misc_json_dict_info(prop, line_number, self.prop_types[prop])
            assert (success)
            if self.prop_types[prop] == "wikibase-item":
                success = self.update_misc_json_dict_info(node2, line_number, None)
                assert (success)

        # update label_json_dict
        if prop in self.label_set:
            success = self.update_misc_json_dict(node1, prop, node2, line_number, rank, "label")
            assert (success)
            return

        # update alias and descriptions
        if prop in self.description_set:
            success = self.update_misc_json_dict(node1, prop, node2, line_number, rank, "description")
            assert (success)
            return

        if prop in self.alias_set:
            success = self.update_misc_json_dict(node1, prop, node2, line_number, rank, "alias")
            assert (success)
            return

            # normal update for claims & qualifiers
        if is_qualifier_edge:
            success = self.update_misc_json_dict(node1, prop, node2, line_number, rank, "qualifier")
        else:
            success = self.update_misc_json_dict(node1, prop, node2, line_number, rank, "statement")

        if (not success):
            if not is_qualifier_edge:
                if self.error_action == 'ignore':
                    pass
                elif self.error_action == 'log':
                    self.warn_log.write(
                        "CORRUPTED_STATEMENT edge at line: [{}] with edge id [{}].\n".format(
                            line_number, e_id))

                    self.corrupted_statement_id = e_id
                elif self.error_action == 'raise':
                    raise KGTKException("CORRUPTED_STATEMENT edge at line: [{}] with edge id [{}].\n".format(
                        line_number, e_id))
                else:
                    raise KGTKException("Unknown error_action {} processing CORRUPTED_STATEMENT edge at line [{}] with "
                                        "edge id [{}].".format(self.error_action, line_number, e_id))
            else:
                if self.error_action == 'ignore':
                    pass
                elif self.error_action == 'log':
                    self.warn_log.write(
                        "CORRUPTED_QUALIFIER edge at line: [{}] with edge id [{}].\n".format(
                            line_number, e_id))
                elif self.error_action == 'raise':
                    raise KGTKException(
                        "CORRUPTED_QUALIFIER edge at line: [{}] with edge id [{}].\n".format(
                            line_number, e_id))
                else:
                    raise KGTKException("Unknown error_action {} processing CORRUPTED_QUALIFIER edge at line [{}] with "
                                        "edge id [{}].".format(self.error_action, line_number, e_id))
        else:
            # success
            if not is_qualifier_edge:
                self.to_append_statement_id = e_id
                self.to_append_statement = [node1, prop]  # path in dictionary for adding future qualifiers

    def init_entity_in_json(self, node: str):
        self.misc_json_dict[node] = {}
        self.misc_json_dict[node]["labels"] = {}
        self.misc_json_dict[node]["descriptions"] = {}
        self.misc_json_dict[node]["aliases"] = {}
        self.misc_json_dict[node]["claims"] = {}
        self.misc_json_dict[node]["sitelinks"] = {}
        if node not in self.prop_types:
            label_type = "item"
        else:
            label_type = "property"
            label_datatype = self.prop_types[node]
            self.misc_json_dict[node]["datatype"] = label_datatype
        self.misc_json_dict[node]["type"] = label_type
        self.misc_json_dict[node]["id"] = node

    def update_misc_json_dict_info(self, node: str, line_number: int, data_type=None):
        if node not in self.misc_json_dict:
            self.init_entity_in_json(node)

        if node.startswith("Q"):
            self.misc_json_dict[node].update(
                {
                    "pageid": -1,
                    "ns": -1,
                    "title": node,
                    "lastrevid": "2000-01-01T00:00:00Z",
                    "type": "item",
                    "id": node}
            )
        elif node.startswith("P"):
            self.misc_json_dict[node].update(
                {
                    "pageid": -1,
                    "ns": -1,
                    "title": "Property:" + node,
                    "lastrevid": "2000-01-01T00:00:00Z",
                    "type": "property",
                    "datatype": data_type,
                    "id": node}
            )
        else:
            if self.error_action == 'ignore':
                pass
            elif self.error_action == 'log':
                self.warn_log.write(
                    "node [{}] at line [{}] is neither an entity nor a property.\n".format(node, line_number))
            elif self.error_action == 'raise':
                raise KGTKException(
                    "node [{}] at line [{}] is neither an entity nor a property.\n".format(node, line_number))
            else:
                raise KGTKException("Unknown error_action {} processing node [{}] at line [{}] is neither an entity "
                                    "nor a property.\n".format(self.error_action, node, line_number))
        return True

    def update_misc_json_dict(self, node1: str, prop: str, node2: str, line_number: int, rank: str, field: str):
        if node1 not in self.misc_json_dict and field != "qualifier":
            self.init_entity_in_json(node1)

        if field == "label":
            label_text, lang = JsonGenerator.process_text_string(node2)
            temp_des_dict = {lang: {"languange": lang, "value": label_text}}
            self.misc_json_dict[node1]["labels"].update(temp_des_dict)
            return True

        if field == "description":
            description_text, lang = JsonGenerator.process_text_string(node2)
            temp_des_dict = {lang: {"languange": lang, "value": description_text}}
            self.misc_json_dict[node1]["descriptions"].update(temp_des_dict)
            return True

        if field == "alias":
            alias_text, lang = JsonGenerator.process_text_string(node2)
            temp_alias_dict = {lang: {"languange": lang, "value": alias_text}}
            if lang in self.misc_json_dict[node1]["aliases"]:
                self.misc_json_dict[node1]["aliases"][lang].append(temp_alias_dict)
            else:
                self.misc_json_dict[node1]["aliases"][lang] = [temp_alias_dict]
            return True

        if field == "statement":
            is_qualifier_edge = False
        elif field == "qualifier":
            is_qualifier_edge = True

        if prop not in self.prop_types:
            if prop in self.wiki_import_prop_types:
                if self.warning:
                    self.warn_log.write(
                        "Property {} created by wikidata json dump at line {} is skipped.\n".format(prop, line_number))
                return True
            else:
                raise KGTKException("property {} at line {} is not defined.".format(prop, line_number))

        if not is_qualifier_edge:
            if prop not in self.misc_json_dict[node1]["claims"]:
                self.misc_json_dict[node1]["claims"][prop] = []
        try:
            if self.prop_types[prop] == "wikibase-item":
                object = self.update_misc_json_dict_item(node1, prop, node2, rank, is_qualifier_edge)
            elif self.prop_types[prop] == "time":
                # print("matched date format yyyy-mm-dd",node1,prop,node2,is_qualifier_edge)
                object = self.update_misc_json_dict_time(node1, prop, node2, rank, is_qualifier_edge)
            elif self.prop_types[prop] == "globe-coordinate":
                object = self.update_misc_json_dict_coordinate(node1, prop, node2, rank, is_qualifier_edge)
            elif self.prop_types[prop] == "quantity":
                object = self.update_misc_json_dict_quantity(node1, prop, node2, rank, is_qualifier_edge)
            elif self.prop_types[prop] == "monolingualtext":
                object = self.update_misc_json_dict_monolingualtext(node1, prop, node2, rank, is_qualifier_edge)
            elif self.prop_types[prop] == "string":
                object = self.update_misc_json_dict_string(node1, prop, node2, rank, is_qualifier_edge)
            elif self.prop_types[prop] == "external-id":
                object = self.update_misc_json_dict_external_id(node1, prop, node2, rank, is_qualifier_edge)
            elif self.prop_types[prop] == "url":
                object = self.update_misc_json_dict_url(node1, prop, node2, rank, is_qualifier_edge)
            else:
                if self.error_action == 'ignore':
                    pass
                elif self.error_action == 'log':
                    self.warn_log.write("property tyepe {} of property {} at line {} is not defined."
                                        .format(self.prop_types[prop], prop, line_number))
                elif self.error_action == 'raise':
                    raise KGTKException("property tyepe {} of property {} at line {} is not defined."
                                        .format(self.prop_types[prop], prop, line_number))
                else:
                    raise KGTKException("Unknown error_action {} processing property tyepe {} of property {} "
                                        "at line {} is not defined."
                                        .format(self.error_action, self.prop_types[prop], prop, line_number))

            if not object:
                if self.warning:
                    self.warn_log.write("edge creation error at line [{}].\n".format(line_number))
                return False

            # process object
            if is_qualifier_edge:
                # update qualifier edge
                if prop in self.misc_json_dict[self.to_append_statement[0]]["claims"] \
                        [self.to_append_statement[1]][-1]["qualifiers"]:
                    self.misc_json_dict[self.to_append_statement[0]]["claims"][self.to_append_statement[1]][-1][
                        "qualifiers"][prop].append(object)
                else:
                    self.misc_json_dict[self.to_append_statement[0]]["claims"][self.to_append_statement[1]][-1][
                        "qualifiers"][prop] = [object]
                if prop not in (
                        self.misc_json_dict[self.to_append_statement[0]]["claims"][self.to_append_statement[1]][-1][
                            "qualifiers-order"]):
                    self.misc_json_dict[self.to_append_statement[0]]["claims"][self.to_append_statement[1]][-1][
                        "qualifiers-order"].append(prop)
            else:
                self.misc_json_dict[node1]["claims"][prop].append(object)
            return True

        except Exception:
            raise KGTKException("illegal edge at line {}.".format(line_number))

    def update_misc_json_dict_item(self, node1: str, prop: str, node2: str, rank: str, is_qualifier_edge: bool):
        if not is_qualifier_edge:
            temp_item_dict = {
                "mainsnak": {
                    "snaktype": "value",
                    "property": prop,
                    "hash": "",
                    "datavalue": {
                        "value": {
                            "entity-type": "item",
                            "numeric-id": 0,
                            "id": node2
                        },
                        "type": "wikibase-entityid"
                    },
                    "datatype": "wikibase-item"
                },
                "type": "statement",
                "id": node1 + prop + node2,
                "rank": rank,  # TODO
                "references": [],
                "qualifiers": {},
                "qualifiers-order": []
            }
        else:
            temp_item_dict = {
                "snaktype": "value",
                "property": prop,
                "hash": "",
                "datavalue": {
                    "value": {
                        "entity-type": "item",
                        "numeric-id": 0,
                        "id": node2
                    },
                    "type": "wikibase-entityid"
                },
                "datatype": "wikibase-item"
            }
        return temp_item_dict

    def update_misc_json_dict_time(self, node1: str, prop: str, node2: str, rank: str, is_qualifier_edge: bool):
        time_string: str
        precision: int
        if self.yyyy_mm_dd_pattern.match(node2):
            time_string = node2 + "-00-00T00:00:00Z"
            precision = 11
        elif self.yyyy_pattern.match(node2):
            time_string = node2 + "-01-01T00:00:00Z"
            precision = 9
        else:
            try:
                precision_str: str
                time_string, precision_str = node2.split("/")
                if time_string.startswith("^"):
                    time_string = time_string[1:]
                if time_string.startswith("+"):
                    time_string = time_string[1:]
                precision = int(precision_str)
            except Exception:
                return None
        if not is_qualifier_edge:
            temp_time_dict = {
                "mainsnak": {
                    "snaktype": "value",
                    "property": prop,
                    "hash": "",
                    "datavalue": {
                        "value": {
                            "time": time_string,
                            "timezone": 0,
                            "before": 0,
                            "after": 0,
                            "precision": precision,
                            "calendarmodel": "http://www.wikidata.org/entity/Q1985727"
                        },
                        "type": "time"
                    },
                    "datatype": "time"
                },
                "type": "statement",
                "id": node1 + prop + node2,
                "rank": rank,  # TODO
                "references": [],
                "qualifiers": {},
                "qualifiers-order": []
            }
        else:
            temp_time_dict = {
                "snaktype": "value",
                "property": prop,
                "hash": "",
                "datavalue": {
                    "value": {
                        "time": time_string,
                        "timezone": 0,
                        "before": 0,
                        "after": 0,
                        "precision": precision,
                        "calendarmodel": "http://www.wikidata.org/entity/Q1985727"
                    },
                    "type": "time"
                },
                "datatype": "time"
            }
        return temp_time_dict

    def update_misc_json_dict_coordinate(self, node1: str, prop: str, node2: str, rank: str, is_qualifier_edge: bool):
        try:
            latitude_str: str
            longitude_str: str
            latitude_str, longitude_str = node2[1:].split("/")
            latitude: float = float(latitude_str)
            longitude: float = float(longitude_str)
        except Exception:
            return None
        if not is_qualifier_edge:
            temp_coordinate_dict = {
                "mainsnak": {
                    "snaktype": "value",
                    "property": prop,
                    "hash": "",
                    "datavalue": {
                        "value": {
                            "latitude": latitude,
                            "longitude": longitude,
                            "altitude": None,
                            "precision": 0.01,  # TODO
                            "globe": "http://www.wikidata.org/entity/Q2"
                        },
                        "type": "globecoordinate"
                    },
                    "datatype": "globecoordinate"
                },
                "type": "statement",
                "id": node1 + prop + node2,
                "rank": rank,
                "references": [],
                "qualifiers": {},
                "qualifiers-order": []
            }
        else:
            temp_coordinate_dict = {
                "snaktype": "value",
                "property": prop,
                "hash": "",
                "datavalue": {
                    "value": {
                        "latitude": latitude,
                        "longitude": longitude,
                        "altitude": None,
                        "precision": 0.01,  # TODO
                        "globe": "http://www.wikidata.org/entity/Q2"
                    },
                    "type": "globecoordinate"
                },
                "datatype": "globecoordinate"
            }
        return temp_coordinate_dict

    def update_misc_json_dict_quantity(self, node1: str, prop: str, node2: str, rank: str, is_qualifier_edge: bool):
        try:
            res = self.quantity_pattern.match(node2).groups()
            amount, lower_bound, upper_bound, unit = res
            amount = JsonGenerator.clean_number_string(amount)
            lower_bound = JsonGenerator.clean_number_string(lower_bound)
            upper_bound = JsonGenerator.clean_number_string(upper_bound)
            unit = "http://www.wikidata.org/entity/" + unit if unit is not None else None
        except Exception:
            return None
        if not is_qualifier_edge:
            temp_quantity_dict = {
                "mainsnak": {
                    "snaktype": "value",
                    "property": prop,
                    "hash": "",
                    "datavalue": {
                        "value": {
                            "amount": amount,
                            "unit": unit,
                            "lowerBound": lower_bound,
                            "UpperBound": upper_bound
                        },
                        "type": "quantity"
                    },
                    "datatype": "quantity"
                },
                "type": "statement",
                "id": node1 + prop + node2,
                "rank": rank,
                "references": [],
                "qualifiers": {},
                "qualifiers-order": []
            }
        else:
            temp_quantity_dict = {
                "snaktype": "value",
                "property": prop,
                "hash": "",
                "datavalue": {
                    "value": {
                        "amount": amount,
                        "unit": unit,
                        "lowerBound": lower_bound,
                        "UpperBound": upper_bound
                    },
                    "type": "quantity"
                },
                "datatype": "quantity"
            }
        return temp_quantity_dict

    def update_misc_json_dict_monolingualtext(self, node1: str, prop: str, node2: str, rank: str,
                                              is_qualifier_edge: bool):
        text_string, lang = JsonGenerator.process_text_string(node2)
        if not is_qualifier_edge:
            temp_mono_dict = {
                "mainsnak": {
                    "snaktype": "value",
                    "property": prop,
                    "hash": "",
                    "datavalue": {
                        "value": {
                            "text": text_string,
                            "language": lang
                        },
                        "type": "monolingualtext"
                    },
                    "datatype": "monolingualtext"
                },
                "type": "statement",
                "id": node1 + prop + node2,
                "rank": rank,
                "references": [],
                "qualifiers": {},
                "qualifiers-order": []
            }
        else:
            temp_mono_dict = {
                "snaktype": "value",
                "property": prop,
                "hash": "",
                "datavalue": {
                    "value": {
                        "text": text_string,
                        "language": lang
                    },
                    "type": "monolingualtext"
                },
                "datatype": "monolingualtext"
            }
        return temp_mono_dict

    def update_misc_json_dict_string(self, node1: str, prop: str, node2: str, rank: str, is_qualifier_edge: bool):
        string, lang = JsonGenerator.process_text_string(node2)
        if not is_qualifier_edge:
            temp_string_dict = {
                "mainsnak": {
                    "snaktype": "value",
                    "property": prop,
                    "hash": "",
                    "datavalue": {"value": string, "type": "string"},
                    "datatype": "string"
                },
                "type": "statement",
                "id": node1 + prop + node2,
                "rank": rank,
                "references": [],
                "qualifiers": {},
                "qualifiers-order": []
            }
        else:
            temp_string_dict = {
                "snaktype": "value",
                "property": prop,
                "hash": "",
                "datavalue": {"value": string, "type": "string"},
                "datatype": "string"
            }
        return temp_string_dict

    def update_misc_json_dict_external_id(self, node1: str, prop: str, node2: str, rank: str, is_qualifier_edge: bool):

        if not is_qualifier_edge:
            temp_e_id_dict = {"mainsnak": {
                "snaktype": "value",
                "property": prop,
                "hash": "",
                "datavalue": {"value": node2, "type": "string"},
                "datatype": "external-id"
            },
                "type": "statement",
                "id": node1 + prop + node2,
                "rank": rank,
                "references": [],
                "qualifiers": {},
                "qualifiers-order": []
            }
        else:
            temp_e_id_dict = {
                "snaktype": "value",
                "property": prop,
                "hash": "",
                "datavalue": {"value": node2, "type": "string"},
                "datatype": "external-id"
            }
        return temp_e_id_dict

    def update_misc_json_dict_url(self, node1: str, prop: str, node2: str, rank: str, is_qualifier_edge: bool):
        if not is_qualifier_edge:
            temp_url_dict = {
                "mainsnak": {
                    "snaktype": "value",
                    "property": prop,
                    "hash": "",
                    "datavalue": {
                        "value": node2,
                        "type": "string"
                    },
                    "datatype": "url"
                },
                "type": "statement",
                "id": node1 + prop + node2,
                "rank": rank,
                "references": [],
                "qualifiers": {},
                "qualifiers-order": []
            }
        else:
            temp_url_dict = {
                "snaktype": "value",
                "property": prop,
                "hash": "",
                "datavalue": {
                    "value": node2,
                    "type": "string"
                },
                "datatype": "url"
            }
        return temp_url_dict

    def set_property(self, node1: str, node2: str, line_num: int):
        try:
            self.prop_types[node1] = self.datatype_mapping[node2.strip()]
        except Exception:
            raise KGTKException(
                "Line {}: DataType [{}] of node [{}] is not supported.\n".format(line_num, node2, node1))

    def read_prop_declaration(self, row: List[str], line_num: int):
        node1, node2, prop, e_id = row[self.node1_idx], row[self.node2_idx], row[self.label_idx], row[self.id_idx]
        if prop == self.property_declaration_label:
            self.set_property(node1, node2, line_num)

    def set_properties(self, prop_file: typing.Optional[Path]):
        self.prop_types: typing.MutableMapping[str, str] = {}
        if prop_file is None:
            return

        if self.verbose:
            print("Reading the properties file %s" % repr(str(prop_file)), file=sys.stderr, flush=True)
        pkr: KgtkReader = KgtkReader.open(prop_file)
        node1_idx: int = pkr.node1_column_idx
        if node1_idx < 0:
            raise KGTKException(
                "The properties file [{}] must have a node1 column (or its alias)".format(str(prop_file)))
        node2_idx: int = pkr.node2_column_idx
        if node2_idx < 0:
            raise KGTKException(
                "The properties file [{}] must have a node2 column (or its alias)".format(str(prop_file)))
        label_idx: int = pkr.label_column_idx
        if self.filter_prop_file and label_idx < 0:
            raise KGTKException(
                "The properties file [{}] must have a label column (or its alias)".format(str(prop_file)))

        prop_count: int = 0
        line_num: int
        row: typing.List[str]
        if self.filter_prop_file:
            for line_num, row in enumerate(pkr):
                prop: str = row[label_idx]
                if prop == self.property_declaration_label:
                    self.set_property(row[node1_idx], row[node2_idx], line_num + 1)
                    prop_count += 1
        else:
            for line_num, row in enumerate(pkr):
                self.set_property(row[node1_idx], row[node2_idx], line_num + 1)
                prop_count += 1

        pkr.close()
        if self.verbose:
            print("Done reading the properties file, %d properties read." % prop_count, file=sys.stderr, flush=True)

    def set_json_dict(self):
        self.misc_json_dict = {}
        # self.label_json_dict = {}
        # self.info_json_dict = {}

    def serialize_to_fp(self, fp) -> int:
        output_lines: int = 0
        for key, value in self.misc_json_dict.items():
            json.dump({key: value}, fp)
            fp.write("\n")
            output_lines += 1
        return output_lines

    def serialize(self):
        '''
        serialize the dictionaries.
        '''
        output_lines: int = 0
        if self.output_prefix == "-":
            if self.verbose:
                print("Serializing to standard output", file=sys.stderr, flush=True)
            output_lines = self.serialize_to_fp(sys.stdout)
        else:
            output_file: str = "{}{}.jsonl".format(self.output_prefix, self.file_num)
            if self.verbose:
                print("Opening the output file %s" % repr(output_file), file=sys.stderr, flush=True)
            with open(output_file, "w") as fp:
                output_lines = self.serialize_to_fp(fp)

        if self.verbose:
            print("Closing the output file, %d lines written." % output_lines, file=sys.stderr, flush=True)
        self.file_num += 1
        self.reset()

    def reset(self):
        self.set_json_dict()
        self.read_num_of_lines = 0
        self.to_append_statement_id = None
        self.to_append_statement = None



from argparse import Action
from enum import Enum

class EnumNameAction(Action):
    """
    Argparse action for handling Enums

    Adapted from:
    https://stackoverflow.com/questions/43968006/support-for-enum-arguments-in-argparse
    """
    def __init__(self, **kwargs):
        # Pop off the type value
        enum = kwargs.pop("type", None)

        # Ensure an Enum subclass is provided
        if enum is None:
            raise ValueError("type must be assigned an Enum when using EnumNameAction")
        if not issubclass(enum, Enum):
            raise TypeError("type must be an Enum when using EnumNameAction")

        # Generate choices from the Enum
        kwargs.setdefault("choices", tuple(e.name for e in enum))

        super().__init__(**kwargs)

        self._enum = enum

    def __call__(self, parser, namespace, values, option_string=None):
        # Convert value back into an Enum
        enum = self._enum[values]
        setattr(namespace, self.dest, enum)

class EnumLowerNameAction(Action):
    """
    Argparse action for handling Enums, presenting the choices in lower case
    but accepting inputs independent of case.  It assumes that the Enum
    member names are upper case.

    Adapted from:
    https://stackoverflow.com/questions/43968006/support-for-enum-arguments-in-argparse
    https://stackoverflow.com/questions/27616778/case-insensitive-argparse-choices

    """
    class cilist(list):
        def __contains__(self, other):
            # Convert to lower case for the comparison, since
            # the list was built in lower case.
            return super().__contains__(other.lower())

    def __init__(self, **kwargs):
        # Pop off the type value
        enum = kwargs.pop("type", None)

        # Ensure an Enum subclass is provided
        if enum is None:
            raise ValueError("type must be assigned an Enum when using EnumNameAction")
        if not issubclass(enum, Enum):
            raise TypeError("type must be an Enum when using EnumNameAction")

        # Generate choices from the Enum, converting the names to lower case:
        kwargs.setdefault("choices", self.cilist(e.name.lower() for e in enum))

        super().__init__(**kwargs)

        self._enum = enum

    def __call__(self, parser, namespace, values, option_string=None):
        # Convert value back into an Enum
        enum = self._enum[values.upper()]
        setattr(namespace, self.dest, enum)

class EnumUpperNameAction(Action):
    """
    Argparse action for handling Enums, presenting the choices in upper case
    but accepting inputs independent of case.  It assumes that the Enum
    member names are upper case.

    Adapted from:
    https://stackoverflow.com/questions/43968006/support-for-enum-arguments-in-argparse
    https://stackoverflow.com/questions/27616778/case-insensitive-argparse-choices

    """
    class cilist(list):
        def __contains__(self, other):
            # Convert to upper case for the comparison, since
            # the list was built in upper case.
            return super().__contains__(other.upper())

    def __init__(self, **kwargs):
        # Pop off the type value
        enum = kwargs.pop("type", None)

        # Ensure an Enum subclass is provided
        if enum is None:
            raise ValueError("type must be assigned an Enum when using EnumNameAction")
        if not issubclass(enum, Enum):
            raise TypeError("type must be an Enum when using EnumNameAction")

        # Generate choices from the Enum, converting the names to upper case:
        kwargs.setdefault("choices", self.cilist(e.name.upper() for e in enum))

        super().__init__(**kwargs)

        self._enum = enum

    def __call__(self, parser, namespace, values, option_string=None):
        # Convert value back into an Enum
        enum = self._enum[values.upper()]
        setattr(namespace, self.dest, enum)



import traceback
from kgtk.exceptions import KGTKException


def parser():
    return {
        'help': 'builds a json lines file to be loaded into elasticsearch from a kgtk edge file.'
    }


def add_arguments(parser):
    """
    Parse Arguments
    Args:
        parser: (argparse.ArgumentParser)

    """

    parser.add_argument('--label-properties', action='store', type=str, dest='label_properties', required=True,
                        help='the name of property which has labels for the node1')

    parser.add_argument('--alias-properties', action='store', type=str, dest='alias_properties', default=None,
                        help='the name of property which has aliases for the node1')

    parser.add_argument('--extra-alias-properties', action='store', type=str, dest='extra_alias_properties',
                        default="P1448,P1705,P1477,P1810",
                        help='comma separated list of properties to be used as additional aliases.')
    # P1448: official name
    # P1705: native label
    # P1477: official name
    # P1810: named as

    parser.add_argument('--description-properties', action='store', type=str, dest='description_properties',
                        default=None,
                        help='the name of property which has descriptions for the node1')

    parser.add_argument('--pagerank-properties', action='store', type=str, dest='pagerank_properties', default=None,
                        help='the name of property which has pagerank for the node1')

    parser.add_argument('--mapping-file', action='store', dest='mapping_file_path', required=True,
                        help='path where a mapping file for the ES index will be output')

    parser.add_argument('--add-text', action='store_true', dest='add_text', default=False,
                        help='add a text field in the json which contains all text in label, alias and description')

    parser.add_argument('--input-file', action='store', dest='input_file_path', required=True,
                        help='input kgtk edge file, sorted by node1')

    parser.add_argument('--output-file', action='store', dest='output_file_path', required=True,
                        help='output json lines file, to be loaded into ES')

    parser.add_argument('--property-datatype-file', action='store', dest='property_datatype_file', default=None,
                        help='A file in KGTK edge file format with data types for properties')
    parser.add_argument('--languages', action='store', type=str, dest='languages',
                        default="en",
                        help='a comma separated list of languages for labels, aliases and descriptions')


def run(**kwargs):
    from kgtk.utils.elasticsearch_manager import ElasticsearchManager
    languages = set(kwargs['languages'].split(","))
    try:

        ElasticsearchManager.build_kgtk_search_input(kwargs['input_file_path'], kwargs['label_properties'],
                                                     kwargs['mapping_file_path'], kwargs['output_file_path'],
                                                     alias_fields=kwargs['alias_properties'],
                                                     extra_alias_properties=kwargs['extra_alias_properties'],
                                                     pagerank_fields=kwargs['pagerank_properties'],
                                                     description_properties=kwargs['description_properties'],
                                                     add_text=kwargs['add_text'],
                                                     property_datatype_file=kwargs['property_datatype_file'],
                                                     languages=languages
                                                     )
    except:
        message = 'Command: build-kgtk-search-input\n'
        message += 'Error Message:  {}\n'.format(traceback.format_exc())
        raise KGTKException(message)
